from collections import Counter
from functools import reduce
import re
from typing import (
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
)

import numpy as np
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass
import scipy.sparse as sp


def tokenize(text: str, sep: str = " ") -> List[str]:
    """Basic tokenizer. Removes punctuation, but not stop words.

    Parameters
    ----------
    text:
        Text to tokenize.
    sep:
        Token separator.

    Examples
    --------
    >>> tokenize("Is this a test? Yes it is.")
    ['is', 'this', 'a', 'test', 'yes', 'it', 'is']
    """
    text = text.lower()
    text = re.sub(r"[\.|,|;|:|!|?|\n]", "", text)
    return text.split(sep)


def extract_ngrams(tokens: List[str], size: int = 1) -> List[str]:
    """Extract ngrams from a list of tokens.

    Parameters
    ----------
    tokens:
        List of tokens.
    size:
        Size of ngrams to extract.

    Examples
    --------
    >>> extract_ngrams(["this", "is", "a", "test"], size=2)
    ['this is', 'is a', 'a test']
    """
    return [
        " ".join(tokens[i : i + size])
        for i in range(0, len(tokens) - size + 1)
    ]


def get_ngram_counts(
    doc: str, ngram_range: Tuple[int, int] = (1, 1)
) -> Counter[str]:
    """Get ngram counts for a document. The ngram range is inclusive.

    Parameters
    ----------
    doc:
        Document to extract ngrams from.
    ngram_range:
        Inclusive range of ngram sizes to extract.

    Examples
    --------
    >>> get_ngram_counts("Red roses red.", ngram_range=(1, 2))
    Counter({'red': 2, 'roses': 1, 'red roses': 1, 'roses red': 1})
    """
    ngram_counts: Counter[str] = Counter()
    tokens = tokenize(doc)
    for size in range(ngram_range[0], ngram_range[1] + 1):
        ngram_counts += Counter(extract_ngrams(tokens, size))
    return ngram_counts


def normalize_csr_rows(X: sp.csr_matrix, norm: str = "l1") -> sp.csr_matrix:
    """Normalize rows of a CSR matrix in place.

    Parameters
    ----------
    X:
        CSR matrix to normalize.
    norm:
        Norm to use for normalization. Either "l1" or "l2".

    Examples
    --------
    >>> X = sp.csr_matrix([[1, 2], [3, 4]], dtype=np.float64)
    >>> normalize_csr_rows(X, norm="l1").toarray()
    array([[0.33333333, 0.66666667],
           [0.42857143, 0.57142857]])
    >>> normalize_csr_rows(X, norm="l2").toarray()
    array([[0.4472136 , 0.89442719],
           [0.6       , 0.8       ]])
    """
    norm_func = {
        "l1": lambda x: np.abs(x).sum(),
        "l2": lambda x: np.sqrt((x**2).sum()),
    }[norm]

    for i in range(X.shape[0]):
        if X[i].sum() == 0.0:
            continue

        X[i, :] /= norm_func(X[i].data)
    return X


@dataclass
class TfidfConfig:
    """Configuration for TfidfVectorizer.

    For more information on tf-idf, see https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfTransformer.html

    Parameters
    ----------
    max_features:
        Maximum number of features to keep. If None, all features are kept.
    ngram_range:
        Inclusive range of ngram sizes to extract.
    smooth_idf:
        Smooth idf weights by adding a constant 1 to the numerator and denominator
        of the idf as if an extra document was seen containing every term once,
        preventing zero divisions.
    vocabulary:
        Vocabulary to use. If None, the vocabulary is inferred from the data.
    norm:
        Normalization to use for the tfidf matrix. Either "l1" or "l2".
    sublinear_tf:
        Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).
    """

    max_features: Optional[int] = None
    ngram_range: Tuple[int, int] = (1, 1)
    smooth_idf: bool = True
    vocabulary: Optional[Dict[str, int]] = None
    norm: Optional[Literal["l1", "l2"]] = None
    sublinear_tf: bool = False


class TfidfVectorizer(BaseModel):
    r"""A simple term frequency-inverse document frequency (tf-idf) vectorizer
    that can be loaded from and serialized to JSON.

    This implementation replicates the behavior of scikit-learn's (as of 1.3.2),
    but only supports a subset of its parameters.

    For more information on tf-idf, see https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfTransformer.html

    Parameters
    ----------
    config:
        Configuration for the vectorizer.
    idf_vector:
        Precomputed idf vector. If None, it is computed from the data.
    vocabulary:
        Vocabulary to use. If None, the vocabulary is inferred from the data.

    Examples
    --------
    >>> docs = ["The quick brown fox", "jumps over", "the lazy dog."]
    >>> vectorizer = TfidfVectorizer(config=TfidfConfig())
    >>> tfidf = vectorizer.fit_transform(docs)
    >>> tfidf.shape
    (3, 8)
    """

    config: TfidfConfig
    idf_vector: List[float] = list()
    vocabulary: Dict[str, int] = Field(default_factory=dict)

    def _get_idf_vector(
        self, ngram_counts: List[Counter[str]], vocab: Dict[str, int]
    ) -> List[float]:
        """Compute the idf vector for the whole corpus from a list of
        ngram counts from each document.

        Parameters
        ----------
        ngram_counts:
            List of ngram counts for each document.
        vocab:
            Vocabulary to use. Each ngram key has an integer value used as the
            column index of the output matrix.
        """
        idf_vector = np.zeros(len(vocab), dtype=np.float64)
        for record in ngram_counts:
            idf_vector[[vocab[t] for t in record.keys() if t in vocab]] += 1
        n_docs = len(ngram_counts) + int(self.config.smooth_idf)
        idf_vector += int(self.config.smooth_idf)
        idf_vector = 1 + np.log(n_docs / (idf_vector))
        return list(idf_vector)

    def _get_tf_matrix(
        self, ngram_counts: List[Counter[str]], vocab: Dict[str, int]
    ) -> sp.csr_matrix:
        """Compute the term frequency matrix for the whole corpus from a
        list of ngram counts from each document.

        Parameters
        ----------
        ngram_counts:
            List of ngram counts for each document (rows of the output matrix).
        vocab:
            Vocabulary to use. Each ngram key has an integer value used as the
            column index of the output matrix.
        """
        tf_matrix = sp.lil_matrix(
            (len(ngram_counts), len(vocab)), dtype=np.float64
        )
        for idx, record in enumerate(ngram_counts):
            pairs = record.items()
            counts = [v for _, v in pairs]
            tf_matrix[idx, [vocab[t] for t, _ in pairs]] = [c for c in counts]
        tf_matrix = tf_matrix.tocsr()
        if self.config.sublinear_tf:
            # applies log in place
            np.log(tf_matrix.data, tf_matrix.data)  # type: ignore
            tf_matrix.data += 1  # type: ignore
        return tf_matrix

    def _get_tfidf(
        self, ngram_counts: List[Counter[str]], vocab: Dict[str, int]
    ) -> sp.csr_matrix:
        """Compute the tfidf matrix over the whole corpus from a list of
        ngram counts from each document.

        Parameters
        ----------
        ngram_counts:
            List of ngram counts for each document.
        vocab:
            Vocabulary to use. Each ngram key has an integer value used as the
            column index of the output matrix.
        """
        tf_matrix: sp.csr_matrix = self._get_tf_matrix(
            ngram_counts, vocab=vocab
        )

        tfidf_matrix = tf_matrix.multiply(np.array(self.idf_vector))  # type: ignore
        return tfidf_matrix.tocsr()  # type: ignore

    def _get_vocabulary(
        self, ngram_counts: Iterable[Counter[str]]
    ) -> dict[str, int]:
        """Get the vocabulary from a list of ngram counts. The vocabulary
        is a mapping from ngrams to integer used as column indices in the
        tfidf matrix.

        Parameters
        ----------
        ngram_counts:
            List of ngram counts for each document.
        """
        counts_corpus = reduce(lambda x, y: x | y, ngram_counts).most_common()
        if self.config.max_features is not None:
            counts_corpus = counts_corpus[: self.config.max_features]
        return {
            t[0]: i
            for i, t in enumerate(sorted(counts_corpus, key=lambda x: x[0]))
        }

    def fit(self, data: Iterable[str]):
        """Fit the vectorizer to a list of documents.

        Parameters
        ----------
        data:
            List of documents contents to fit the vectorizer to."""
        counts_records: List[Counter[str]] = [
            get_ngram_counts(doc, self.config.ngram_range) for doc in data
        ]
        vocab = self.config.vocabulary or self._get_vocabulary(counts_records)
        self.idf_vector = self._get_idf_vector(counts_records, vocab=vocab)
        self.vocabulary = vocab

    def transform(self, data: Iterable[str]) -> sp.csr_matrix:
        """Transform a list of documents into a tfidf matrix.
        The model must be fit before calling this method.

        Parameters
        ----------
        data:
            List of documents contents to transform.
        """
        if not self.vocabulary:
            raise ValueError("Vocabulary is empty. Call `fit` first.")
        counts_records = [
            get_ngram_counts(doc, self.config.ngram_range) for doc in data
        ]
        counts_records = [
            Counter({k: v for k, v in doc.items() if k in self.vocabulary})
            for doc in counts_records
        ]
        tfidf = self._get_tfidf(counts_records, vocab=self.vocabulary)
        if self.config.norm is not None:
            return normalize_csr_rows(tfidf, norm=self.config.norm)
        return tfidf

    def fit_transform(self, data: Iterable[str]) -> sp.csr_matrix:
        """Fit the vectorizer to a list of documents and transform them
        into a tfidf matrix.

        Parameters
        ----------
        data:
            List of documents contents to fit the vectorizer to and transform.
        """
        self.fit(list(data))
        return self.transform(data)



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from io import BytesIO
import re
from typing import List, Optional, Set
import yaml
from rdflib.term import URIRef
from rdflib import Graph, URIRef, Literal
from rdflib.namespace import RDF
from gimie import logger
from gimie.graph.namespaces import SDO, MD4I
from gimie.parsers.abstract import Parser
from gimie.utils.uri import is_valid_orcid, extract_doi_match


class CffParser(Parser):
    """Parse DOI and authors from CITATION.cff."""

    def __init__(self, subject: str):
        super().__init__(subject)

    def parse(self, data: bytes) -> Graph:
        """Extracts a DOI link and list of authors from a CFF file and returns a
        graph with a single triple <subject> <schema:citation> <doi>
        and a number of author objects with <schema:name> and <md4i:orcid> values.
        If no DOI is found, it will not be included in the graph.
        If no authors are found, it will not be included in the graph.
        If neither authors nor DOI are found, an empty graph is returned.
        """
        extracted_cff_triples = Graph()
        doi = get_cff_doi(data)
        authors = get_cff_authors(data)

        if doi:
            extracted_cff_triples.add(
                (self.subject, SDO.citation, URIRef(doi))
            )
        if not authors:
            return extracted_cff_triples
        for author in authors:
            if is_valid_orcid(author["orcid"]):
                orcid = URIRef(author["orcid"])
                extracted_cff_triples.add(
                    (self.subject, SDO.author, URIRef(orcid))
                )
                extracted_cff_triples.add(
                    (
                        URIRef(orcid),
                        SDO.name,
                        Literal(
                            author["given-names"]
                            + " "
                            + author["family-names"]
                        ),
                    )
                )
                extracted_cff_triples.add(
                    (
                        orcid,
                        MD4I.orcidId,
                        Literal(orcid),
                    )
                )
                extracted_cff_triples.add(
                    (
                        orcid,
                        SDO.affiliation,
                        Literal(author["affiliation"]),
                    )
                )
                extracted_cff_triples.add((orcid, RDF.type, SDO.Person))
        return extracted_cff_triples


def doi_to_url(doi: str) -> str:
    """Formats a doi to an https URL to doi.org.

    Parameters
    ----------
    doi
        doi where the scheme (e.g. https://) and
        hostname (e.g. doi.org) may be missing.

    Returns
    -------
    str
        doi formatted as a valid url. Base url
        is set to https://doi.org when missing.

    Examples
    --------
    >>> doi_to_url("10.0000/example.abcd")
    'https://doi.org/10.0000/example.abcd'
    >>> doi_to_url("doi.org/10.0000/example.abcd")
    'https://doi.org/10.0000/example.abcd'
    >>> doi_to_url("https://doi.org/10.0000/example.abcd")
    'https://doi.org/10.0000/example.abcd'
    """

    doi_match = extract_doi_match(doi)

    if doi_match is None:
        raise ValueError(f"Not a valid DOI: {doi}")

    return f"https://doi.org/{doi_match}"


def get_cff_doi(data: bytes) -> Optional[str]:
    """Given a CFF file, returns the DOI, if any.

    Parameters
    ----------
    data
        The cff file body as bytes.

    Returns
    -------
    str, optional
        doi formatted as a valid url

    Examples
    --------
    >>> get_cff_doi(bytes("doi:   10.5281/zenodo.1234", encoding="utf8"))
    'https://doi.org/10.5281/zenodo.1234'
    >>> get_cff_doi(bytes("abc: def", encoding="utf8"))

    """

    try:
        cff = yaml.safe_load(data.decode())
    except yaml.scanner.ScannerError:
        logger.warning("cannot read CITATION.cff, skipped.")
        return None

    try:
        doi_url = doi_to_url(cff["doi"])
    # No doi in cff file
    except (KeyError, TypeError):
        logger.warning("CITATION.cff does not contain a 'doi' key.")
        doi_url = None
    # doi is malformed
    except ValueError as err:
        logger.warning(err)
        doi_url = None

    return doi_url


def get_cff_authors(data: bytes) -> Optional[List[dict[str, str]]]:
    """Given a CFF file, returns a list of dictionaries containing orcid, affiliation, first and last names of authors, if any.

    Parameters
    ----------
    data
        The cff file body as bytes.

    Returns
    -------
    list(dict), optional
        orcid, names strings of authors

    """

    try:
        cff = yaml.safe_load(data.decode())
    except yaml.scanner.ScannerError:
        logger.warning("cannot read CITATION.cff, skipped.")
        return None

    authors = []
    try:
        for author in cff["authors"]:
            author_dict = {
                "family-names": author.get("family-names", ""),
                "given-names": author.get("given-names", ""),
                "orcid": author.get("orcid", ""),
                "affiliation": author.get("affiliation", ""),
            }
            authors.append(author_dict)
    except KeyError:
        logger.warning("CITATION.cff does not contain an 'authors' key.")
        return None

    return authors if authors else None



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Git providers from which metadata can be extracted by gimie."""
from typing import Dict, Optional, Type
from gimie.extractors.abstract import Extractor
from gimie.extractors.github import GithubExtractor
from gimie.extractors.gitlab import GitlabExtractor
from gimie.extractors.git import GitExtractor
from gimie.utils.uri import validate_url

GIT_PROVIDERS: Dict[str, Type[Extractor]] = {
    "git": GitExtractor,
    "github": GithubExtractor,
    "gitlab": GitlabExtractor,
}


def get_extractor(
    url: str,
    source: str,
    base_url: Optional[str] = None,
    local_path: Optional[str] = None,
) -> Extractor:
    """Instantiate the correct extractor for a given source.

    Parameters
    -----------
    URL
        Where the repository metadata is extracted from.
    source
        The source of the repository (git, gitlab, github, ...).
    base_url
        The base URL of the git remote.
    local_path
        If applicable, the path to the directory where the
        repository is located.

    Examples
    --------
    >>> extractor = get_extractor(
    ...     "https://github.com/sdsc-ordes/gimie",
    ...     "github"
    ... )
    """
    try:
        return GIT_PROVIDERS[source](
            url, base_url=base_url, local_path=local_path
        )
    except KeyError as err:
        raise ValueError(
            f"Unknown git provider: {source}.\n"
            f"Supported sources: {', '.join(GIT_PROVIDERS)}"
        ) from err


def infer_git_provider(url: str) -> str:
    """Given a git repository URL, return the corresponding git provider.
    Local path or unsupported git providers will return "git".

    Examples
    --------
    >>> infer_git_provider("https://gitlab.com/foo/bar")
    'gitlab'
    >>> infer_git_provider("/foo/bar")
    'git'
    >>> infer_git_provider("https://codeberg.org/dnkl/foot")
    'git'
    """
    # Fall back to git if local path
    if not validate_url(url):
        return "git"

    # NOTE: We just check if the provider name is in the URL.
    # We may want to use a more robust check.
    for name in GIT_PROVIDERS.keys():
        if name in url and name != "git":
            return name

    # Fall back to git for unsupported providers
    return "git"



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Files which can be parsed by gimie."""
from pathlib import Path
from typing import Iterable, NamedTuple, Optional, Set, Type

from gimie.graph import Property
from gimie.io import Resource
from gimie.parsers.abstract import Parser
from gimie.parsers.license import LicenseParser, is_license_filename
from gimie.parsers.cff import CffParser

from rdflib import Graph


class ParserInfo(NamedTuple):
    default: bool
    type: Type[Parser]


PARSERS = {
    "license": ParserInfo(default=True, type=LicenseParser),
    "cff": ParserInfo(default=True, type=CffParser),
}


def get_parser(name: str) -> Type[Parser]:
    """Get a parser by name."""
    parser = PARSERS.get(name, None)
    if parser is None:
        raise ValueError(
            f"Unknown parser: {name}.\n"
            f"Supported parsers: {', '.join(PARSERS)}"
        )
    return parser.type


def list_default_parsers() -> Set[str]:
    """List the names of all default parsers."""
    return {k for k, v in PARSERS.items() if v.default}


def list_parsers() -> Set[str]:
    """List the names of all parsers."""
    return set(PARSERS.keys())


def select_parser(
    path: Path,
    parsers: Optional[Set[str]] = None,
) -> Optional[Type[Parser]]:
    """Select the appropriate parser from a collection based on a file path.
    If no parser is found, return None.

    Parameters
    ----------
    path:
        The path of the file to parse.
    parsers:
        A set of parser names. If None, use the default collection.
    """
    # Only parse licenses and citations in the root directory
    if is_license_filename(path.name) and len(path.parts) == 1:
        name = "license"
    elif path.name == "CITATION.cff" and len(path.parts) == 1:
        name = "cff"
    else:
        return None

    if name not in (parsers or list_parsers()):
        return None
    return get_parser(name)


def parse_files(
    subject: str,
    files: Iterable[Resource],
    parsers: Optional[Set[str]] = None,
) -> Graph:
    """For each input file, select appropriate parser among a collection and
    parse its contents. Return the union of all parsed properties in the form of triples.
    If no parser is found for a given file, skip it.

    Parameters
    ----------
    subject:
        The subject URI of the repository.
    files:
        A collection of file-like objects.
    parsers:
        A set of parser names. If None, use the default collection.
    """
    parsed_properties = Graph()
    for file in files:
        parser = select_parser(file.path, parsers)
        if not parser:
            continue
        data = file.open().read()
        parsed_properties |= parser(subject).parse(data or b"")
    return parsed_properties



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from dataclasses import dataclass
import os
import requests
from datetime import datetime
from dateutil.parser import isoparse
from functools import cached_property
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse
from dotenv import load_dotenv
from gimie.io import RemoteResource
from gimie.models import (
    Organization,
    Person,
    Repository,
)
from gimie.extractors.abstract import Extractor
from gimie.extractors.common.queries import send_graphql_query, send_rest_query

load_dotenv()


@dataclass
class GitlabExtractor(Extractor):
    """Extractor for Gitlab repositories. Uses the Gitlab GraphQL API to
    extract metadata into linked data.
    url: str
        The url of the git repository.
    base_url: Optional[str]
        The base url of the git remote.

    """

    url: str
    base_url: Optional[str] = None
    local_path: Optional[str] = None

    token: Optional[str] = None

    def list_files(self) -> List[RemoteResource]:
        """takes the root repository folder and returns the list of files present"""
        file_list = []
        file_dict = self._repo_data["repository"]["tree"]["blobs"]["nodes"]
        defaultbranchref = self._repo_data["repository"]["rootRef"]
        for item in file_dict:
            file = RemoteResource(
                path=item["name"],
                url=f'{self.url}/-/raw/{defaultbranchref}/{item["name"]}',
                headers=self._headers,
            )
            file_list.append(file)
        return file_list

    def extract(self) -> Repository:
        """Extract metadata from target Gitlab repository."""

        # fetch metadata
        data = self._repo_data

        # NOTE(identifier): Each Gitlab project has a unique identifier (integer)
        # NOTE(author): Fetches only the group directly related to the project
        # the group takes the form: parent/subgroup

        # NOTE(contributors): contributors = project members
        # who are not owners + those that have written merge requests
        # owners are either multiple individuals or a group. If no user
        # is marked as owner, contributors are project members or merge
        # request authors
        repo_meta = dict(
            authors=self._safe_extract_author(data),
            contributors=self._safe_extract_contributors(data),
            date_created=isoparse(data["createdAt"][:-1]),
            date_modified=isoparse(data["lastActivityAt"][:-1]),
            description=data["description"],
            identifier=urlparse(data["id"]).path.split("/")[2],
            keywords=data["topics"],
            name=self.path,
            prog_langs=[lang["name"] for lang in data["languages"]],
            url=self.url,
        )

        if data["releases"]["edges"]:
            repo_meta["date_published"] = isoparse(
                data["releases"]["edges"][0]["node"]["releasedAt"]
            )

        if data["releases"] and (len(data["releases"]["edges"]) > 0):
            # go into releases and take the name from the first node (most recent)
            version = data["releases"]["edges"][0]["node"]["name"]
            repo_meta["version"] = version
            repo_meta[
                "download_url"
            ] = f"{self.url}/-/archive/{version}/{self.path.split('/')[-1]}-{version}.tar.gz"
        return Repository(**repo_meta)  # type: ignore

    def _safe_extract_author(
        self, repo: Dict[str, Any]
    ) -> List[Union[Person, Organization]]:
        """Extract the author from a GraphQL repository node.
        projectMembers is used if available, otherwise the author
        is inferred from the project url."""
        members = repo["projectMembers"]["edges"]
        if len(members) > 0:
            owners = filter(
                lambda m: m["node"]["accessLevel"]["stringValue"] == "OWNER",
                members,
            )
            return [
                self._get_author(owner["node"]["user"]) for owner in owners
            ]

        if repo["group"] is not None:
            return [self._get_author(repo["group"])]

        # If the author is absent from the GraphQL response (permission bug),
        # fallback to the REST API
        return [self._user_from_rest(self.path.split("/")[0])]

    def _safe_extract_contributors(
        self, repo: dict[str, Any]
    ) -> List[Person] | None:
        members = [
            user["node"]["user"]
            for user in repo["projectMembers"]["edges"]
            if user["node"]["accessLevel"]["stringValue"] != "OWNER"
        ]
        merge_request_authors = [
            author["node"]["author"]
            for author in repo["mergeRequests"]["edges"]
        ]
        contributors = members + merge_request_authors
        # Drop duplicate (unhashable) dicts by "id" key
        uniq_contrib = list({c["id"]: c for c in contributors}.values())
        return [self._get_user(contrib) for contrib in uniq_contrib]

    @cached_property
    def _repo_data(self) -> Dict[str, Any]:
        """Fetch repository metadata from GraphQL endpoint."""
        data = {"path": self.path}
        project_query = """
        query project_query($path: ID!) {
            project(fullPath: $path) {
                name
                id
                description
                createdAt
                lastActivityAt
                group {
                    id
                    name
                    description
                    avatarUrl
                    webUrl
                }
                languages {
                    name
                    share
                }
                topics
                projectMembers {
                    edges {
                        node {
                        id
                        accessLevel {
                            stringValue
                        }
                        user {
                            id
                            name
                            username
                            publicEmail
                            webUrl
                        }
                        }
                    }
                }
                mergeRequests{
                    edges {
                    node {
                        author {
                        id
                        name
                        username
                        publicEmail
                        webUrl
                        }
                    }
                    }
                }
                repository {
                    rootRef
                    tree{
                        blobs{
                            nodes {
                                name
                                webUrl
                            }
                        }
                    }
                }
                releases {
                    edges {
                    node {
                        name
                        releasedAt
                    }
                    }
                }
        }
        }
        """
        response = send_graphql_query(
            self.graphql_endpoint, project_query, data, self._headers
        )
        if "errors" in response:
            raise ValueError(response["errors"])

        return response["data"]["project"]

    @cached_property
    def _headers(self) -> Any:
        """Set authentication headers for Gitlab API requests."""
        try:
            if not self.token:
                self.token = os.environ.get("GITLAB_TOKEN")
                assert self.token
            headers = {"Authorization": f"token {self.token}"}

            login = requests.get(f"{self.rest_endpoint}/user", headers=headers)
            assert login.json().get("login")
        except AssertionError:
            return {}
        else:
            return headers

    def _get_author(self, node: Dict[str, Any]) -> Union[Organization, Person]:
        """Given the GraphQL node for a repository owner,
        return the author as a Person or Organization object."""
        # Is this the best test?
        if "username" in node:
            return self._get_user(node)
        return self._get_organization(node)

    def _get_organization(self, node: Dict[str, Any]) -> Organization:
        """Extract details from a GraphQL organization node."""
        return Organization(
            _id=node["webUrl"],
            name=node["name"],
            description=node.get("description"),
            logo=node.get("avatarUrl"),
        )

    def _get_user(self, node: Dict[str, Any]) -> Person:
        """Extract details from a GraphQL user node."""
        return Person(
            _id=node["webUrl"],
            identifier=node["username"],
            name=node.get("name"),
            email=node.get("publicEmail"),
        )

    def _user_from_rest(self, username: str) -> Person:
        """Given a username, use the REST API to retrieve the Person object."""

        author = send_rest_query(
            self.rest_endpoint,
            f"/users?username={username}",
            self._headers,
        )
        if isinstance(author, list):
            author = author[0]

        return Person(
            _id=author["web_url"],
            identifier=author["username"],
            name=author.get("name"),
        )

    @property
    def rest_endpoint(self) -> str:
        return f"{self.base}/api/v4/"

    @property
    def graphql_endpoint(self) -> str:
        return f"{self.base}/api"



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions used throughout gimie."""

from typing import List, Literal
from urllib.parse import urlparse
import re

from gimie.graph.namespaces import GIMIE


def validate_url(url: str):
    """Checks if input is a valid URL.
    credits: https://stackoverflow.com/a/38020041

    Examples
    -------------
    >>> validate_url('/data/my_repo')
    False
    >>> validate_url(532)
    False
    >>> validate_url('https://www.github.com/sdsc-ordes/gimie')
    True
    >>> validate_url('github.com/sdsc-ordes/gimie')
    False
    """
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except AttributeError:
        return False


def generate_uri(ref: str):
    """Given a reference (e.g. commit sha), return a URI.

    Parameters
    ----------
    path:
        Path to the repository, either local or a URL.


    Returns
    -------
    fair_uri:
        A unique resource identifier (URI) for the repository path.

    Examples
    --------
    >>> generate_uri("abc")
    'https://sdsc-ordes.github.io/gimie/abc'
    """
    return str(GIMIE[ref])


def is_valid_orcid(orcid):
    """Check if the input is a valid ORCID according to definition from orcid.org [1]_.
    .. [1] [https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier](https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier)

    Parameters
    ----------
    orcid:
        The ORCID to validate.

    Returns
    -------
    bool:
        True if the ORCID is valid, False otherwise.

    Examples
    --------
    >>> is_valid_orcid("https://orcid.org/0000-0001-2345-6789")
    True
    >>> is_valid_orcid("0000-0001-2345-6789")
    False
    >>> is_valid_orcid("http://orcid.org/0000-0001-2345-6789")
    False

    """
    return bool(
        re.match(
            r"(https:\/\/)?orcid.org\/\d{4}-\d{4}-\d{4}-\d{4}", str(orcid)
        )
    )


def extract_doi_match(doi):
    """Extracts doi from the input if it contains a valid DOI according to definition from crossref.org [1]_.
    .. [1] [https://www.crossref.org/blog/dois-and-matching-regular-expressions](https://www.crossref.org/blog/dois-and-matching-regular-expressions)

    Parameters
    ----------
    doi:
        The DOI to validate.

    Returns
    -------
    str:
        The extracted short DOI if it is valid, None otherwise.

    Examples
    --------
    >>> extract_doi_match("10.5281/zenodo.1234567")
    '10.5281/zenodo.1234567'
    >>> extract_doi_match("https://doi.org/10.5281/zenodo.1234567")
    '10.5281/zenodo.1234567'
    """
    match = re.search(
        r"10.\d{4,9}/[-._;()/:A-Z0-9]+$", doi, flags=re.IGNORECASE
    )
    if match:
        return match.group()



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests
from typing import Any, Dict, List, Union


def send_rest_query(
    api: str, query: str, headers: Dict[str, str]
) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    """Generic function to send a query to the GitHub/GitLab rest API."""
    resp = requests.get(
        url=f"{api}/{query}",
        headers=headers,
    )

    if resp.status_code != 200:
        error_msg = resp.json().get("message", "")
        if "API rate limit exceeded" in error_msg:
            raise ConnectionError(
                "Authentication failed: API rate limit exceeded. Please check that you have added "
                "your GITHUB_TOKEN or GITLAB_TOKEN to your environment variables."
            )
        raise ConnectionError(f"API request failed: {error_msg}")
    return resp.json()


def send_graphql_query(
    api: str, query: str, data: Dict[str, Any], headers: Dict[str, str]
) -> Dict[str, Any]:
    """Generic function to send a GraphQL query to the GitHub/GitLab API."""
    resp = requests.post(
        url=f"{api}/graphql",
        json={
            "query": query,
            "variables": data,
        },
        headers=headers,
    )

    if resp.status_code != 200:
        error_msg = resp.json().get("message", "")
        if "API rate limit exceeded" in error_msg:
            raise ConnectionError(
                "Authentication failed: API rate limit exceeded. Please check that you have added "
                "your GITHUB_TOKEN or GITLAB_TOKEN to your environment variables."
            )
        raise ConnectionError(f"API request failed: {error_msg}")
    return resp.json()



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
from io import BytesIO
import pkgutil
import re
from typing import List, Optional, Set

import numpy as np
import scipy.sparse as sp
from rdflib.term import URIRef
from rdflib import Graph
from gimie.graph.namespaces import SDO
from gimie.parsers.abstract import Parser, Property
from gimie.utils.text_processing import TfidfVectorizer


class LicenseParser(Parser):
    """Parse LICENSE body into schema:license <spdx-url>.
    Uses tf-idf-based matching."""

    def __init__(self, subject: str):
        super().__init__(subject)

    def parse(self, data: bytes) -> Graph:
        """Extracts an spdx URL from a license file and returns a
        graph with a single triple <url> <schema:license> <spdx_url>.
        If no matching URL is found, an empty graph is returned.
        """
        license_facts = Graph()
        license_url = match_license(data)

        if license_url:
            license_facts.add((self.subject, SDO.license, URIRef(license_url)))
        return license_facts


def match_license(data: bytes, min_similarity: float = 0.9) -> Optional[str]:
    """Given a license file, returns the url of the most similar spdx license.
    This is done using TF-IDF on the license text and getting the
    closest match in the SPDX license corpus based on cosine similarity.

    Parameters
    ----------
    data:
        The license body as bytes.

    Examples
    --------
    >>> match_license(open('LICENSE', 'rb').read())
    'https://spdx.org/licenses/Apache-2.0.html'
    """
    # Compute tfidf vector for input license
    vectorizer = load_tfidf_vectorizer()
    input_vec = vectorizer.transform([data.decode()])

    # Load ids and tfidf vectors for spdx licenses
    spdx_licenses = load_spdx_ids()
    spdx_vecs = load_tfidf_matrix()
    # Compute cosine similarity between input_vec and spdx vectors
    sim: np.ndarray = (input_vec * spdx_vecs.T).todense()
    # Pick the most similar spdx vector
    closest_idx = np.argmax(sim)
    # If similarity is below threshold, return None
    if sim[0, closest_idx] < min_similarity:
        return None
    closest_id = spdx_licenses[closest_idx]
    return f"https://spdx.org/licenses/{closest_id}.html"


def load_tfidf_vectorizer() -> TfidfVectorizer:
    """Load tfidf matrix and vectorizer from disk."""

    data = pkgutil.get_data(__name__, "data/tfidf_vectorizer.json")
    if data is None:
        raise FileNotFoundError("Could not find tfidf_vectorizer.json")
    return TfidfVectorizer.model_validate_json(data)


def load_spdx_ids() -> List[str]:
    """Load spdx licenses from disk."""
    data = pkgutil.get_data(__name__, "data/spdx_licenses.csv")
    if data is None:
        raise FileNotFoundError("Could not find spdx_licenses.csv")
    reader = csv.reader(data.decode().split("\n"))
    return [l[0] for l in reader if l]


def load_tfidf_matrix() -> sp.csr_matrix:
    """Load pre-computed tfidf matrix of spdx licenses from disk.
    Matrix has dimensions (n_licenses, n_features)."""
    data = pkgutil.get_data(__name__, "data/tfidf_matrix.npz")
    if data is None:
        raise FileNotFoundError("Could not find tfidf_matrix.npz")
    return sp.load_npz(BytesIO(data))


def is_license_filename(filename: str) -> bool:
    """Given an input filename, returns a boolean indicating whether the filename path looks like a license.

    Parameters
    ----------
    filename:
        A filename to check.

    Examples
    --------
    >>> is_license_filename('LICENSE-APACHE')
    True
    >>> is_license_filename('README.md')
    False
    """
    if filename.startswith("."):
        return False
    pattern = r".*(license(s)?.*|lizenz|reus(e|ing).*|copy(ing)?.*)(\.(txt|md|rst))?$"
    if re.match(pattern, filename, flags=re.IGNORECASE):
        return True
    return False



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract for Git repository extractors."""
from abc import ABC, abstractmethod
from typing import List, Optional

from urllib.parse import urlparse

from gimie.io import Resource
from gimie.models import Repository


class Extractor(ABC):
    """Extractor is an Abstract Base Class. It is only meant
    to define a standard interface for all git repository extractors.

    Subclasses for different git providers must implement
    extract() and list_files() methods.
    """

    def __init__(
        self,
        url: str,
        base_url: Optional[str] = None,
        local_path: Optional[str] = None,
    ):
        self.url = url
        self.base_url = base_url
        self.local_path = local_path

    @abstractmethod
    def extract(self) -> Repository:
        """Extract metadata from the git provider into a Repository object."""
        ...

    @abstractmethod
    def list_files(self) -> List[Resource]:
        """List all files in the repository HEAD."""
        ...

    @property
    def path(self) -> str:
        """Path to the repository without the base URL."""
        if self.base_url is None:
            return urlparse(self.url).path.strip("/")
        return self.url.removeprefix(self.base_url).strip("/")

    @property
    def base(self) -> str:
        """Base URL of the remote."""
        if self.base_url is None:
            url = urlparse(self.url)
            return f"{url.scheme}://{url.netloc}"
        return self.base_url



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line interface to the gimie package."""
from enum import Enum
from typing import List, Optional

import click
import typer

from gimie import __version__
from gimie.parsers import get_parser, list_default_parsers, list_parsers
from gimie.project import Project

app = typer.Typer(add_completion=False)


# Used to autogenerate docs with sphinx-click
@click.group()
def cli():
    """Command line group"""
    pass


class RDFFormatChoice(str, Enum):
    ttl = "ttl"
    jsonld = "json-ld"
    nt = "nt"


def version_callback(value: bool):
    if value:
        print(f"gimie {__version__}")
        # Exits successfully
        raise typer.Exit()


@app.command()
def data(
    url: str,
    format: RDFFormatChoice = typer.Option(
        RDFFormatChoice.ttl,
        "--format",
        show_choices=True,
        help="Output serialization format for the RDF graph.",
    ),
    base_url: Optional[str] = typer.Option(
        None,
        "--base-url",
        help="Specify the base URL of the git provider. Inferred by default.",
    ),
    include_parser: Optional[List[str]] = typer.Option(
        None,
        "--include-parser",
        "-I",
        help="Only include selected parser. Use 'gimie parsers' to list parsers.",
    ),
    exclude_parser: Optional[List[str]] = typer.Option(
        None,
        "--exclude-parser",
        "-X",
        help="Exclude selected parser.",
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        help="Display version and exit",
        callback=version_callback,
    ),
):
    """Extract linked metadata from a Git repository at the target URL.

    The output is sent to stdout, and turtle is used as the default serialization format.
    """
    parser_names = list_default_parsers()
    if exclude_parser:
        parser_names -= set([parser for parser in exclude_parser])
    if include_parser:
        parser_names = set([parser for parser in include_parser])
    proj = Project(url, base_url=base_url, parser_names=parser_names)
    repo_meta = proj.extract()
    print(repo_meta.serialize(format=format.value))


@app.command()
def advice(url: str):
    """Show a metadata completion report for a Git repository
    at the target URL.

    NOTE: Not implemented yet"""
    ...
    raise typer.Exit()


@app.command()
def parsers(
    verbose: bool = typer.Option(
        False, "--verbose", help="Show parser description."
    )
):
    """List available parsers, specifying which are default.
    If --verbose is used, show parser description."""
    message = ""
    parsers = list_parsers()
    default_parsers = list_default_parsers()

    for name in parsers:
        # Each parser gets their name in bold green
        title = typer.style(name, fg=typer.colors.GREEN, bold=True)
        default = " (default)" if name in default_parsers else ""
        description = f" - {get_parser(name).__doc__}" if verbose else ""

        parser_line = f"{title}{default}{description}"
        message += f"{parser_line}\n"

    typer.echo(message)


typer_cli = typer.main.get_command(app)
cli.add_command(typer_cli, "cli")


# This callback is triggered when gimie is called without subcommand
@app.callback()
def callback(
    version: Optional[bool] = typer.Option(
        None, "--version", callback=version_callback
    )
):
    """gimie digs Git repositories for metadata."""


if __name__ == "__main__":
    app()



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data models to represent nodes in the graph generated by gimie."""
from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
import datetime
from typing import List, Optional, Union

from calamus.schema import JsonLDSchema
from calamus import fields
from rdflib import Graph

from gimie.graph.namespaces import SDO


@dataclass(order=True)
class Release:
    """
    This class represents a release of a repository.

    Parameters
    ----------
    tag: str
        The tag of the release.
    date: datetime.datetime
        The date of the release.
    commit_hash: str
        The commit hash of the release.
    """

    tag: str = field(compare=False)
    date: datetime = field(compare=True)
    commit_hash: str = field(compare=False)


@dataclass
class Organization:
    """See http//schema.org/Organization"""

    _id: str
    name: str
    legal_name: Optional[str] = None
    email: Optional[List[str]] = None
    description: Optional[str] = None
    logo: Optional[str] = None


class OrganizationSchema(JsonLDSchema):
    _id = fields.Id()
    name = fields.String(SDO.name)
    legal_name = fields.String(SDO.legalName)
    email = fields.String(SDO.email)
    description = fields.String(SDO.description)
    logo = fields.IRI(SDO.logo)

    class Meta:
        rdf_type = SDO.Organization
        model = Organization


@dataclass
class Person:
    """See http//schema.org/Person"""

    _id: str
    identifier: str
    name: Optional[str] = None
    email: Optional[str] = None
    affiliations: Optional[List[Organization]] = None

    def __str__(self):
        name = f"({self.name}) " if self.name else ""
        email = f"<{self.email}> " if self.email else ""
        orgs = (
            f"[{', '.join([org.name for org in self.affiliations])}]"
            if self.affiliations
            else ""
        )
        return f"{self.identifier} {name}{email}{orgs}".strip(" ")


class PersonSchema(JsonLDSchema):
    _id = fields.Id()
    identifier = fields.String(SDO.identifier)
    name = fields.String(SDO.name)
    affiliations = fields.Nested(
        SDO.affiliation, OrganizationSchema, many=True
    )

    class Meta:
        rdf_type = SDO.Person
        model = Person


@dataclass
class Repository:
    """This class represents a git repository.
    It does not contain any information about the content of the repository.
    See https://schema.org/SoftwareSourceCode
    """

    url: str
    name: str

    authors: Optional[List[Union[Organization, Person]]] = None
    contributors: Optional[List[Person]] = None
    date_created: Optional[datetime] = None
    date_modified: Optional[datetime] = None
    date_published: Optional[datetime] = None
    description: Optional[str] = None
    download_url: Optional[str] = None
    identifier: Optional[str] = None
    keywords: Optional[List[str]] = None
    licenses: Optional[List[str]] = None
    parent_repository: Optional[str] = None
    prog_langs: Optional[List[str]] = None
    version: Optional[str] = None

    @property
    def _id(self) -> str:
        """Unique identifier for the repository."""
        return self.url

    def to_graph(self) -> Graph:
        """Convert repository to RDF graph."""
        jd = RepositorySchema().dumps(self)
        g: Graph = Graph().parse(format="json-ld", data=str(jd))
        g.bind("schema", SDO)
        return g

    def serialize(self, format: str = "ttl", **kwargs) -> str:
        """Serialize the RDF graph representing the instance."""
        return self.to_graph().serialize(format=format, **kwargs)  # type: ignore

    def jsonld(self) -> str:
        """Alias for jsonld serialization."""
        return self.serialize(format="json-ld")


class RepositorySchema(JsonLDSchema):
    """This defines the schema used for json-ld serialization."""

    _id = fields.Id()
    authors = fields.Nested(
        SDO.author, [PersonSchema, OrganizationSchema], many=True
    )
    contributors = fields.Nested(SDO.contributor, PersonSchema, many=True)
    date_created = fields.Date(SDO.dateCreated)
    date_modified = fields.Date(SDO.dateModified)
    date_published = fields.Date(SDO.datePublished)
    description = fields.String(SDO.description)
    download_url = fields.IRI(SDO.downloadUrl)
    identifier = fields.String(SDO.identifier)
    keywords = fields.List(SDO.keywords, fields.String)
    licenses = fields.List(SDO.license, fields.IRI)
    name = fields.String(SDO.name)
    parent_repository = fields.IRI(SDO.isBasedOn)
    prog_langs = fields.List(SDO.programmingLanguage, fields.String)
    url = fields.IRI(SDO.codeRepository)
    version = fields.String(SDO.version)

    class Meta:
        rdf_type = SDO.SoftwareSourceCode
        model = Repository
        add_value_types = False



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extractor which uses a locally available (usually cloned) repository."""
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
import os
import shutil
import tempfile
from typing import List, Optional
import uuid

import git
import pydriller

from gimie.io import LocalResource
from gimie.models import Person, Repository
from gimie.extractors.abstract import Extractor
from pathlib import Path


@dataclass
class GitExtractor(Extractor):
    """
    This class is responsible for extracting metadata from a git repository.

    Parameters
    ----------
    url: str
        The url of the git repository.
    base_url: Optional[str]
        The base url of the git remote.
    local_path: Optional[str]
        The local path where the cloned git repository is located.

    Attributes
    ----------
    uri: Optional[str]
        The URI to assign the repository in RDF.
    repository: Repository
        The repository we are extracting metadata from.
    """

    url: str
    base_url: Optional[str] = None
    local_path: Optional[str] = None
    _cloned: bool = False

    def extract(self) -> Repository:
        # Assuming author is the first person to commit
        self.repository = self._repo_data

        repo_meta = dict(
            authors=[self._get_creator()],
            contributors=self._get_contributors(),
            date_created=self._get_creation_date(),
            date_modified=self._get_modification_date(),
            name=self.path,
            url=self.url,
        )

        return Repository(**repo_meta)  # type: ignore

    def list_files(self) -> List[LocalResource]:
        self.repository = self._repo_data
        file_list = []

        for path in Path(self.local_path).rglob("*"):  # type: ignore
            if (path.parts[0] == ".git") or not path.is_file():
                continue
            file_list.append(LocalResource(path))

        return file_list

    def __del__(self):
        """Cleanup the cloned repo if it was cloned and is located in tempdir."""
        try:
            # Can't be too careful with temp files
            tempdir = tempfile.gettempdir()
            if (
                self.local_path
                and self._cloned
                and self.local_path.startswith(tempdir)
                and tempdir != os.getcwd()
            ):
                shutil.rmtree(self.local_path)
        except AttributeError:
            pass

    @cached_property
    def _repo_data(self) -> pydriller.Repository:
        """Get the repository data by accessing local data or cloning."""
        if self.local_path is None:
            self._cloned = True
            self.local_path = tempfile.TemporaryDirectory().name
            git.Repo.clone_from(self.url, self.local_path)  # type: ignore
        return pydriller.Repository(self.local_path)

    def _get_contributors(self) -> List[Person]:
        """Get the authors of the repository."""
        authors = set()
        for commit in self.repository.traverse_commits():
            if commit.author is not None:
                authors.add((commit.author.name, commit.author.email))
        return [self._dev_to_person(name, email) for name, email in authors]

    def _get_creation_date(self) -> Optional[datetime]:
        """Get the creation date of the repository."""
        try:
            return next(self.repository.traverse_commits()).author_date
        except StopIteration:
            return None

    def _get_modification_date(self) -> Optional[datetime]:
        """Get the last modification date of the repository."""
        commit = None
        try:
            for commit in self.repository.traverse_commits():
                pass
        except (StopIteration, NameError):
            pass
        finally:
            return commit.author_date if commit else None

    def _get_creator(self) -> Optional[Person]:
        """Get the creator of the repository."""
        try:
            creator = next(self.repository.traverse_commits()).author
            return self._dev_to_person(creator.name, creator.email)
        except StopIteration:
            return None

    def _dev_to_person(
        self, name: Optional[str], email: Optional[str]
    ) -> Person:
        """Convert a Developer object to a Person object."""
        if name is None:
            uid = str(uuid.uuid4())
        else:
            uid = name.replace(" ", "_").lower()
        dev_id = f"{self.url}/{uid}"
        return Person(
            _id=dev_id,
            identifier=uid,
            name=name,
            email=email,
        )



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from functools import reduce
from typing import Iterable, Set
from rdflib import Graph, URIRef
from gimie.graph import Property


class Parser(ABC):
    """
    Parser is an Abstract Base Class. It is only meant
    to define a standard interface for all parsers.

    All subclasses must implement parse(). A parser parses
    bytes data into a set of predicate-object tuples.

    Parameters
    ----------
    subject:
        The subject of a triple (subject - predicate - object) to be used for writing parsed properties to.
    """

    def __init__(self, subject: str):
        self.subject = URIRef(subject)

    @abstractmethod
    def parse(self, data: bytes) -> Graph:
        """Extract rdf graph from a source."""
        ...

    def parse_all(self, docs: Iterable[bytes]) -> Graph:
        """Parse multiple sources and return the union of
        triples."""

        properties = map(self.parse, docs)
        return reduce(lambda p1, p2: p1 | p2, properties)



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Orchestration of multiple extractors for a given project.
This is the main entry point for end-to-end analysis."""
from typing import Iterable, Optional, Tuple

from rdflib import Graph
from rdflib.term import URIRef
from urllib.parse import urlparse

from gimie.extractors import get_extractor, infer_git_provider
from gimie.graph.operations import properties_to_graph
from gimie.parsers import parse_files
from gimie.utils.uri import validate_url


class Project:
    """A class to represent a project's git repository.


    Parameters
    ----------
    path:
        The full path (URL) of the repository.
    base_url:
        The base URL of the git remote. Can be used to
        specify delimitation between base URL and project name.
    git_provider:
        The name of the git provider to extract metadata from.
        ('git', 'github', 'gitlab')
    parser_names:
        Names of file parsers to use. ('license').
        If None, default parsers are used (see gimie.parsers.PARSERS).

    Examples
    --------
    >>> proj = Project("https://github.com/sdsc-ordes/gimie")
    >>> assert isinstance(proj.extract(), Graph)
    """

    def __init__(
        self,
        path: str,
        base_url: Optional[str] = None,
        git_provider: Optional[str] = None,
        parser_names: Optional[Iterable[str]] = None,
    ):
        if not git_provider:
            git_provider = infer_git_provider(path)

        self.base_url = base_url
        self.project_dir = None
        self._cloned = False
        if validate_url(path):
            self.url = path
        else:
            self.project_dir = path

        self.extractor = get_extractor(
            self.url,
            git_provider,
            base_url=self.base_url,
            local_path=self.project_dir,
        )
        if parser_names:
            self.parsers = set(parser_names)
        else:
            self.parsers = None

    def extract(self) -> Graph:
        """Extract repository metadata from git provider to RDF graph and enrich with
        metadata parsed from file contents."""

        repo = self.extractor.extract()
        repo_graph = repo.to_graph()

        files = self.extractor.list_files()
        parsed_graph = parse_files(self.url, files, self.parsers)

        repo_graph += parsed_graph
        return repo_graph


def split_git_url(url: str) -> Tuple[str, str]:
    """Split a git URL into base URL and project path.

    Examples
    --------
    >>> split_git_url("https://gitlab.com/foo/bar")
    ('https://gitlab.com', 'foo/bar')
    """
    base_url = urlparse(url).scheme + "://" + urlparse(url).netloc
    project = urlparse(url).path.strip("/")
    return base_url, project



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Operations on graphs."""
from functools import reduce
from typing import Set

from rdflib import Graph
from rdflib.term import URIRef

from gimie.graph import Property


def combine_graphs(*graphs: Graph) -> Graph:
    """Combines an arbitrary number of input graphs
    into a single graph."""
    return reduce(lambda g1, g2: g1 | g2, graphs)


def properties_to_graph(uri: URIRef, properties: Set[Property]) -> Graph:
    """Attaches a set of predicate-object tuples to input
    URI to produce an RDF graph."""
    g = Graph()
    for pred, obj in properties:
        g.add((uri, pred, obj))
    return g



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Standard input interfaces to local or remote resources for gimie."""

import io
import os
from pathlib import Path
import requests
from typing import Iterator, Optional, Union


class Resource:
    """Abstract class for read-only access to local or remote resources via
    a file-like interface.

    Parameters
    ----------
    path:
        The local relative path to the resource.
    """

    path: Path

    def open(self) -> io.RawIOBase:
        raise NotImplementedError


class LocalResource(Resource):
    """Providing read-only access to local data via a file-like interface.

    Examples
    --------
    >>> resource = LocalResource("README.md")
    """

    def __init__(self, path: Union[str, os.PathLike]):
        self.path: Path = Path(path)

    def open(self) -> io.RawIOBase:
        return io.FileIO(self.path, mode="r")


class RemoteResource(Resource):
    """Provides read-only access to remote data via a file-like interface.

    Parameters
    ----------
    url:
        The URL where the resource. can be downladed from.
    headers:
        Optional headers to pass to the request.

    Examples
    --------
    >>> url = "https://raw.githubusercontent.com/sdsc-ordes/gimie/main/README.md"
    >>> content = RemoteResource("README.md", url).open().read()
    >>> assert isinstance(content, bytes)
    """

    def __init__(self, path: str, url: str, headers: Optional[dict] = None):
        self.path = Path(path)
        self.url = url
        self.headers = headers or {}

    def open(self) -> io.RawIOBase:
        resp = requests.get(
            self.url, headers=self.headers, stream=True
        ).iter_content(chunk_size=128)
        return IterStream(resp)


class IterStream(io.RawIOBase):
    """Wraps an iterator under a like a file-like interface.
    Empty elements in the iterator are ignored.

    Parameters
    ----------
    iterator:
        An iterator yielding bytes.

    Examples
    --------
    >>> stream = IterStream(iter([b"Hello ", b"", b"World"]))
    >>> stream.read()
    b'Hello World'
    """

    def __init__(self, iterator: Iterator[bytes]):
        self.leftover = b""
        self.iterator = iterator

    def readable(self):
        return True

    def readinto(self, b):
        try:
            l = len(b)  # We're supposed to return at most this much
            while True:
                chunk = self.leftover or next(self.iterator)
                # skip empty elements
                if not chunk:
                    continue
                output, self.leftover = chunk[:l], chunk[l:]
                b[: len(output)] = output
                return len(output)
        except StopIteration:
            return 0  # indicate EOF
